import hashlib
import uuid


def gen_infra_name() -> str:
    random_uuid = uuid.uuid4().hex
    hash_object = hashlib.sha256(random_uuid.encode())
    hash_value = hash_object.hexdigest()[:20]
    return f"flyte-k8sdsinfra-{hash_value}"



from kubernetes import client


def cleanup_resources(resources: client.V1ResourceRequirements):
    filtered_resources_limits = {k: v for k, v in resources.limits.items() if not is_zero_like(v)}
    filtered_resources_requests = {k: v for k, v in resources.requests.items() if not is_zero_like(v)}
    resources.limits = filtered_resources_limits
    resources.requests = filtered_resources_requests


# We have noticed that for the fields that are not specified, there will be 0 or "0" filled in
# this is helper function to filter those values out.
def is_zero_like(value):
    return (
        value is None
        or value == "0"
        or value == 0
        or str(value).strip().lower() == "0"
        or str(value).strip().lower() == "0gb"
    )


def convert_flyte_to_k8s_fields(resources_dict):
    return {("memory" if k == "mem" else k): v for k, v in resources_dict.items()}



import pytest
import asyncio
import unittest
from unittest.mock import patch, MagicMock
from kubernetes import client
from kubernetes.client.rest import ApiException
from flytekitplugins.k8sdataservice.sensor import CleanupSensor


@pytest.mark.asyncio
@patch("flytekitplugins.k8sdataservice.sensor.KubeConfig")
@patch("flytekitplugins.k8sdataservice.sensor.client.AppsV1Api")
@patch("flytekitplugins.k8sdataservice.sensor.client.CoreV1Api")
@patch("flytekitplugins.k8sdataservice.sensor.client.CustomObjectsApi")
@patch("flytekitplugins.k8sdataservice.sensor.logger")
async def test_poke_no_cleanup(
    mock_logger,
    mock_custom_api,
    mock_core_v1_api,
    mock_apps_v1_api,
    mock_kube_config
):
    mock_kube_config.return_value = MagicMock()
    sensor = CleanupSensor(name="test-sensor")
    await asyncio.create_task(
        sensor.poke(release_name="test-release",
                    cleanup_data_service=False,
                    cluster="test-cluster")
    )
    mock_kube_config.return_value.load_kube_config.assert_called_once()
    assert isinstance(sensor.apps_v1_api, MagicMock)
    assert isinstance(sensor.core_v1_api, MagicMock)
    assert isinstance(sensor.custom_api, MagicMock)
    assert sensor.release_name == "test-release"
    assert sensor.cleanup_data_service is False
    assert sensor.namespace == "flyte"
    assert sensor.cluster == "test-cluster"


@pytest.mark.asyncio
@patch("flytekitplugins.k8sdataservice.sensor.CleanupSensor.delete_data_service")
@patch("flytekitplugins.k8sdataservice.sensor.KubeConfig")
@patch("flytekitplugins.k8sdataservice.sensor.client.AppsV1Api")
@patch("flytekitplugins.k8sdataservice.sensor.client.CoreV1Api")
@patch("flytekitplugins.k8sdataservice.sensor.client.CustomObjectsApi")
@patch("flytekitplugins.k8sdataservice.sensor.logger")
async def test_poke_with_cleanup(
    mock_logger,
    mock_custom_api,
    mock_core_v1_api,
    mock_apps_v1_api,
    mock_kube_config,
    mock_delete_data_service
):
    mock_kube_config.return_value = MagicMock()
    # Initialize CleanupSensor instance
    sensor = CleanupSensor(name="test-sensor")

    async def sensor_poke_task():
        await sensor.poke(
            release_name="test-release",
            cleanup_data_service=True,
            cluster="test-cluster"
        )

    # Use asyncio.wait_for for timeout handling (compatible with older Python versions)
    try:
        await asyncio.wait_for(sensor_poke_task(), timeout=30)
    except asyncio.TimeoutError:
        pytest.fail("Test timed out")

    # Assertions for logger and delete_data_service call
    mock_logger.info.assert_any_call(
        "The training job is in terminal stage, deleting graph engine test-release"
    )
    mock_delete_data_service.assert_called_once()


class TestCleanupSensor(unittest.TestCase):
    @patch("flytekitplugins.k8sdataservice.sensor.logger")
    @patch("flytekitplugins.k8sdataservice.sensor.client.CoreV1Api")
    @patch("flytekitplugins.k8sdataservice.sensor.client.AppsV1Api")
    def test_delete_data_service(self, mock_apps_v1_api, mock_core_v1_api, mock_logger):
        # Configure mocks for service deletion
        mock_core_v1_api_instance = mock_core_v1_api.return_value
        mock_apps_v1_api_instance = mock_apps_v1_api.return_value
        # Initialize sensor and set required properties
        sensor = CleanupSensor(name="test-sensor")
        sensor.core_v1_api = mock_core_v1_api_instance
        sensor.apps_v1_api = mock_apps_v1_api_instance
        sensor.release_name = "test-release"
        sensor.namespace = "test-namespace"

        # Call delete_data_service
        sensor.delete_data_service()

        # Verify the service deletion calls
        mock_core_v1_api_instance.delete_namespaced_service.assert_called_once_with(
            name="test-release", namespace="test-namespace", body=client.V1DeleteOptions()
        )
        mock_apps_v1_api_instance.delete_namespaced_stateful_set.assert_called_once_with(
            name="test-release", namespace="test-namespace", body=client.V1DeleteOptions()
        )
        # Check for logger messages indicating success
        mock_logger.info.assert_any_call("Deleted Service: test-release")
        mock_logger.info.assert_any_call("Deleted StatefulSet: test-release")

    @patch("flytekitplugins.k8sdataservice.sensor.logger")
    @patch("flytekitplugins.k8sdataservice.sensor.client.CoreV1Api")
    @patch("flytekitplugins.k8sdataservice.sensor.client.AppsV1Api")
    def test_delete_data_service_with_exceptions(self, mock_apps_v1_api, mock_core_v1_api, mock_logger):
        # Configure mocks to raise exceptions
        mock_core_v1_api_instance = mock_core_v1_api.return_value
        mock_apps_v1_api_instance = mock_apps_v1_api.return_value

        mock_core_v1_api_instance.delete_namespaced_service.side_effect = ApiException("Service deletion failed")
        mock_apps_v1_api_instance.delete_namespaced_stateful_set.side_effect = ApiException("StatefulSet deletion failed")

        # Initialize sensor and set required properties
        sensor = CleanupSensor(name="test-sensor")
        sensor.core_v1_api = mock_core_v1_api_instance
        sensor.apps_v1_api = mock_apps_v1_api_instance
        sensor.release_name = "test-release"
        sensor.namespace = "test-namespace"

        # Call delete_data_service and handle exceptions
        sensor.delete_data_service()

        # Verify that errors were logged
        service_error_logged = any("Error deleting Service" in call[0][0] for call in mock_logger.error.call_args_list)
        statefulset_error_logged = any("Error deleting StatefulSet" in call[0][0] for call in mock_logger.error.call_args_list)

        self.assertTrue(service_error_logged, "Service deletion error not logged as expected")
        self.assertTrue(statefulset_error_logged, "StatefulSet deletion error not logged as expected")



import uuid

from flytekitplugins.k8sdataservice.k8s.kube_config import KubeConfig
from kubernetes import client
from kubernetes.client.rest import ApiException
from utils.resources import cleanup_resources, convert_flyte_to_k8s_fields

from flytekit import logger

APPNAME = "service-name"
DEFAULT_RESOURCES = client.V1ResourceRequirements(
    requests={"cpu": "2", "memory": "10G"}, limits={"cpu": "6", "memory": "16G"}
)


class K8sManager:
    def __init__(self):
        self.config = KubeConfig()
        self.config.load_kube_config()
        self.apps_v1_api = client.AppsV1Api()
        self.core_v1_api = client.CoreV1Api()

    def set_configs(self, data_service_config):
        self.data_service_config = data_service_config
        self.labels = {}
        self.namespace = "flyte"
        self.name = None
        self.name = data_service_config.get("Name", None)
        if self.name is None:
            self.name = f"k8s-dataservice-{uuid.uuid4().hex[:8]}"

    def create_data_service(self) -> str:
        svc_name = self.create_service()
        logger.info(f"Created service: {svc_name}")
        stateful_set_obj = self.create_stateful_set_object()
        name = self.create_stateful_set(stateful_set_obj)
        return name

    def create_stateful_set(self, stateful_set_object) -> str:
        api_response = None
        try:
            api_response = self.apps_v1_api.create_namespaced_stateful_set(
                namespace=self.namespace, body=stateful_set_object
            )
            logger.info(f"Created statefulset in K8s API server: {api_response}")
        except ApiException as e:
            logger.error(f"Exception when calling AppsV1Api->create_namespaced_stateful_set: {e}\n")
            raise
        return api_response.metadata.name

    def create_stateful_set_object(self):
        container = self._create_container()
        template = self._create_pod_template(container)
        spec = self._create_stateful_set_spec(template)
        return client.V1StatefulSet(
            api_version="apps/v1",
            kind="StatefulSet",
            metadata=client.V1ObjectMeta(
                labels=self.labels,
                name=self.name,
                annotations={},
            ),
            spec=spec,
        )

    def _create_container(self):
        ss_replicas = self.data_service_config.get("Replicas", 1)
        port = self.data_service_config.get("Port", 40000)
        ss_env = [
            client.V1EnvVar(name="GE_BASE_PORT", value=str(port)),
            client.V1EnvVar(name="GE_COUNT", value=str(int(ss_replicas))),
            client.V1EnvVar(name="SERVER_PORT", value=str(port)),
        ]
        return client.V1Container(
            name=self.name,
            image=self.data_service_config["Image"],
            image_pull_policy="IfNotPresent",
            ports=[client.V1ContainerPort(container_port=port, name="graph-engine")],
            command=self.data_service_config["Command"],
            env=ss_env,
            resources=self.get_resources(),
        )

    def _create_pod_template(self, container):
        self.labels.update({"app.kubernetes.io/instance": self.name})
        return client.V1PodTemplateSpec(
            metadata=client.V1ObjectMeta(
                labels=self.labels,
                annotations={},
            ),
            spec=client.V1PodSpec(
                containers=[container],
                security_context=client.V1PodSecurityContext(
                    fs_group=1001,
                    run_as_group=1001,
                    run_as_non_root=True,
                    run_as_user=1001,
                ),
            ),
        )

    def _create_stateful_set_spec(self, template):
        ss_replicas = self.data_service_config.get("Replicas", 1)
        return client.V1StatefulSetSpec(
            replicas=int(ss_replicas),
            selector=client.V1LabelSelector(
                match_labels={"app.kubernetes.io/instance": self.name},
            ),
            service_name=self.name,
            template=template,
        )

    def create_service(self) -> str:
        namespace = self.namespace
        logger.info(f"creating a service at namespace {namespace} with name {self.name}")
        port = self.data_service_config.get("Port", 40000)
        self.labels.update({"app.kubernetes.io/instance": self.name, "app": APPNAME})
        body = client.V1Service(
            api_version="v1",
            kind="Service",
            metadata=client.V1ObjectMeta(
                name=self.name,
                labels=self.labels,
                namespace=namespace,
            ),
            spec=client.V1ServiceSpec(
                selector={"app.kubernetes.io/instance": self.name},
                type="ClusterIP",
                ports=[
                    client.V1ServicePort(
                        port=port,
                        target_port=port,
                        name=self.name,
                    )
                ],
            ),
        )
        logger.info(
            f"Service configuration in namespace {namespace} and name {self.name} is completed, posting request to K8s API server..."
        )
        api_response = None
        try:
            api_response = self.core_v1_api.create_namespaced_service(namespace=namespace, body=body)
        except ApiException as e:
            logger.error(f"Exception when calling CoreV1Api->create_namespaced_service: {e}")
            raise e
        # This will not happen in K8s API, but in case.
        if api_response is None or not hasattr(api_response, "metadata") or not hasattr(api_response.metadata, "name"):
            raise ValueError("Invalid response from Kubernetes API - missing metadata or name")
        return api_response.metadata.name

    def check_stateful_set_status(self, name) -> str:
        try:
            stateful_set = self.apps_v1_api.read_namespaced_stateful_set(name=name, namespace=self.namespace)
            status = stateful_set.status
            logger.info(f"StatefulSet status: {status}")
            conditions = status.conditions if status and status.conditions else []
            logger.info(f"StatefulSet conditions: {conditions}")

            if status.replicas == 0:
                logger.info(
                    f"StatefulSet {name} is pending. replicas: {status.replicas}, available: {status.available_replicas }"
                )
                return "pending"

            if status.replicas > 0 and (
                status.replicas == status.available_replicas or status.replicas == status.ready_replicas
            ):
                logger.info(
                    f"StatefulSet {name} has succeeded. replicas: {status.replicas}, available: {status.available_replicas }"
                )
                return "success"

            if status.replicas > 0 and status.available_replicas is not None and status.available_replicas >= 0:
                logger.info(
                    f"StatefulSet {name} is running. replicas: {status.replicas}, available: {status.available_replicas }"
                )
                return "running"

            logger.info(
                f"StatefulSet {name} status is unknown. Replicas:  {status.replicas}, available: {status.available_replicas }"
            )
            return "failed"
        except ApiException as e:
            logger.error(f"Exception when calling AppsV1Api->read_namespaced_stateful_set: {e}")
            return f"Error checking status of StatefulSet {name}: {e}"

    def delete_stateful_set(self, name: str):
        try:
            self.apps_v1_api.delete_namespaced_stateful_set(
                name=name, namespace=self.namespace, body=client.V1DeleteOptions()
            )
            logger.info(f"Deleted StatefulSet: {name}")
        except ApiException as e:
            logger.error(f"Exception when calling AppsV1Api->delete_namespaced_stateful_set: {e}")

    def delete_service(self, name: str):
        try:
            self.core_v1_api.delete_namespaced_service(
                name=name, namespace=self.namespace, body=client.V1DeleteOptions()
            )
            logger.info(f"Deleted Service: {name}")
        except ApiException as e:
            logger.error(f"Exception when calling CoreV1Api->delete_namespaced_service: {e}")

    def get_resources(self) -> client.V1ResourceRequirements:
        res = DEFAULT_RESOURCES
        flyteidl_limits = self.data_service_config.get("Limits", None)
        flyteidl_requests = self.data_service_config.get("Requests", None)
        logger.info(f"Flyte Resources: limits: {flyteidl_limits} and requests {flyteidl_requests}")
        if flyteidl_limits is not None:
            res.limits = convert_flyte_to_k8s_fields(flyteidl_limits)
            logger.info(f"Resources limits updated is: {res.limits}")
        if flyteidl_requests is not None:
            res.requests = convert_flyte_to_k8s_fields(flyteidl_requests)
            logger.info(f"Resources requests updated is: {res.requests}")
        cleanup_resources(res)
        logger.info(f"Resources cleaned up is: {res}")
        return res



from datetime import timedelta
from unittest.mock import patch
from google.protobuf import json_format
from flytekitplugins.k8sdataservice.task import DataServiceConfig
from flytekitplugins.k8sdataservice.connector import DataServiceMetadata
from google.protobuf.struct_pb2 import Struct
from flyteidl.core.execution_pb2 import TaskExecution
import flytekit.models.interface as interface_models
from flytekit.extend.backend.base_connector import ConnectorRegistry
from flytekit.interfaces.cli_identifiers import Identifier
from flytekit.models import literals, task, types
from flytekit.models.core.identifier import ResourceType
from flytekit.models.task import TaskTemplate


cmd = ["command", "args"]


def create_test_task_metadata() -> task.TaskMetadata:
    return task.TaskMetadata(
       discoverable= True,
        runtime=task.RuntimeMetadata(task.RuntimeMetadata.RuntimeType.FLYTE_SDK, "1.0.0", "python"),
        timeout=timedelta(days=1),
        retries=literals.RetryStrategy(3),
        interruptible=True,
        discovery_version="0.1.1b0",
        deprecated_error_message="This is deprecated!",
        cache_serializable=True,
        pod_template_name="A",
        cache_ignore_input_vars=(),
    )


def create_test_setup(original_name: str = "gnn-1234", existing_release_name: str = "gnn-2345"):
    task_id = Identifier(
        resource_type=ResourceType.TASK, project="project", domain="domain", name="name", version="version"
    )
    task_metadata = create_test_task_metadata()
    s = Struct()
    if existing_release_name != "":
        s.update({
            "Name": original_name,
            "Image": "image",
            "Command": cmd,
            "Cluster": "ei-dev2",
            "ExistingReleaseName": existing_release_name,
        })
    else:
        s.update({
            "Name": original_name,
            "Image": "image",
            "Command": cmd,
            "Cluster": "ei-dev2",
        })
    task_config = json_format.MessageToDict(s)
    return task_id, task_metadata, task_config


@patch("flytekitplugins.k8sdataservice.connector.K8sManager.create_data_service", return_value="gnn-1234")
@patch("flytekitplugins.k8sdataservice.connector.K8sManager.check_stateful_set_status", return_value="succeeded")
@patch("flytekitplugins.k8sdataservice.connector.K8sManager.delete_stateful_set")
@patch("flytekitplugins.k8sdataservice.connector.K8sManager.delete_service")
def test_gnn_connector(mock_delete_service, mock_delete_stateful_set, mock_check_status, mock_create_data_service):
    connector = ConnectorRegistry.get_connector("dataservicetask")
    task_id, task_metadata, task_config = create_test_setup(existing_release_name="")
    int_type = types.LiteralType(types.SimpleType.INTEGER)
    interfaces = interface_models.TypedInterface(
        {
            "a": interface_models.Variable(int_type, "description1"),
            "b": interface_models.Variable(int_type, "description2"),
        },
        {},
    )
    task_inputs = literals.LiteralMap(
        {
            "a": literals.Literal(scalar=literals.Scalar(primitive=literals.Primitive(integer=1))),
            "b": literals.Literal(scalar=literals.Scalar(primitive=literals.Primitive(integer=1))),
        },
    )

    dummy_template = TaskTemplate(
        id=task_id,
        custom=task_config,
        metadata=task_metadata,
        interface=interfaces,
        type="dataservicetask",
    )

    expected_resource_metadata = DataServiceMetadata(
            dataservice_config=DataServiceConfig(Name="gnn-1234", Image="image", Command=cmd, Cluster="ei-dev2"),
            name="gnn-1234")
    # Test create method
    res_resource_metadata = connector.create(dummy_template, task_inputs)
    assert res_resource_metadata == expected_resource_metadata
    mock_create_data_service.assert_called_once()

    # Test get method
    res = connector.get(res_resource_metadata)
    assert res.phase == TaskExecution.SUCCEEDED
    assert res.outputs.get("data_service_name") == 'gnn-1234'
    mock_check_status.assert_called_once_with("gnn-1234")

    # # Test delete method
    connector.delete(res_resource_metadata)
    mock_delete_stateful_set.assert_called_once_with("gnn-1234")
    mock_delete_service.assert_called_once_with("gnn-1234")


@patch("flytekitplugins.k8sdataservice.connector.K8sManager.create_data_service", return_value="gnn-1234")
@patch("flytekitplugins.k8sdataservice.connector.K8sManager.check_stateful_set_status", return_value="succeeded")
@patch("flytekitplugins.k8sdataservice.connector.K8sManager.delete_stateful_set")
@patch("flytekitplugins.k8sdataservice.connector.K8sManager.delete_service")
def test_gnn_connector_reuse_data_service(mock_delete_service, mock_delete_stateful_set, mock_check_status, mock_create_data_service):
    connector = ConnectorRegistry.get_connector("dataservicetask")
    task_id, task_metadata, task_config = create_test_setup(original_name="gnn-2345", existing_release_name="gnn-2345")

    int_type = types.LiteralType(types.SimpleType.INTEGER)
    interfaces = interface_models.TypedInterface(
        {
            "a": interface_models.Variable(int_type, "description1"),
            "b": interface_models.Variable(int_type, "description2"),
        },
        {},
    )
    task_inputs = literals.LiteralMap(
        {
            "a": literals.Literal(scalar=literals.Scalar(primitive=literals.Primitive(integer=1))),
            "b": literals.Literal(scalar=literals.Scalar(primitive=literals.Primitive(integer=1))),
        },
    )

    dummy_template = TaskTemplate(
        id=task_id,
        custom=task_config,
        metadata=task_metadata,
        interface=interfaces,
        type="dataservicetask",
    )

    expected_resource_metadata = DataServiceMetadata(
            dataservice_config=DataServiceConfig(
                Name="gnn-2345", Image="image", Command=cmd, Cluster="ei-dev2", ExistingReleaseName="gnn-2345"),
            name="gnn-2345")

    # Test create method, and create_data_service should have not been called
    res_resource_metadata = connector.create(dummy_template, task_inputs)
    assert res_resource_metadata == expected_resource_metadata
    mock_create_data_service.assert_not_called()

    # Test get method
    res = connector.get(res_resource_metadata)
    assert res.phase == TaskExecution.SUCCEEDED
    mock_check_status.assert_called_once_with("gnn-2345")

    # # Test delete method
    connector.delete(res_resource_metadata)
    mock_delete_stateful_set.assert_called_once_with("gnn-2345")
    mock_delete_service.assert_called_once_with("gnn-2345")


@patch("flytekitplugins.k8sdataservice.connector.K8sManager.create_data_service", return_value="gnn-1234")
@patch("flytekitplugins.k8sdataservice.connector.K8sManager.check_stateful_set_status", return_value="running")
@patch("flytekitplugins.k8sdataservice.connector.K8sManager.delete_stateful_set")
@patch("flytekitplugins.k8sdataservice.connector.K8sManager.delete_service")
def test_gnn_connector_status(mock_delete_service, mock_delete_stateful_set, mock_check_status, mock_create_data_service):
    connector = ConnectorRegistry.get_connector("dataservicetask")
    task_id, task_metadata, task_config = create_test_setup(original_name="gnn-2345", existing_release_name="gnn-2345")

    int_type = types.LiteralType(types.SimpleType.INTEGER)
    interfaces = interface_models.TypedInterface(
        {
            "a": interface_models.Variable(int_type, "description1"),
            "b": interface_models.Variable(int_type, "description2"),
        },
        {},
    )
    task_inputs = literals.LiteralMap(
        {
            "a": literals.Literal(scalar=literals.Scalar(primitive=literals.Primitive(integer=1))),
            "b": literals.Literal(scalar=literals.Scalar(primitive=literals.Primitive(integer=1))),
        },
    )

    dummy_template = TaskTemplate(
        id=task_id,
        custom=task_config,
        metadata=task_metadata,
        interface=interfaces,
        type="dataservicetask",
    )

    expected_resource_metadata = DataServiceMetadata(
            dataservice_config=DataServiceConfig(
                Name="gnn-2345", Image="image", Command=cmd, Cluster="ei-dev2", ExistingReleaseName="gnn-2345"),
            name="gnn-2345")
    # Test create method, and create_data_service should have not been called
    res_resource_metadata = connector.create(dummy_template, task_inputs)
    assert res_resource_metadata == expected_resource_metadata
    mock_create_data_service.assert_not_called()

    # Test get method
    res = connector.get(res_resource_metadata)
    assert res.phase == TaskExecution.RUNNING
    mock_check_status.assert_called_once_with("gnn-2345")

    # # Test delete methods are not called
    connector.delete(res_resource_metadata)
    mock_delete_stateful_set.assert_called_once_with("gnn-2345")
    mock_delete_service.assert_called_once_with("gnn-2345")


@patch("flytekitplugins.k8sdataservice.connector.K8sManager.create_data_service", return_value="gnn-1234")
@patch("flytekitplugins.k8sdataservice.connector.K8sManager.check_stateful_set_status", return_value="succeeded")
@patch("flytekitplugins.k8sdataservice.connector.K8sManager.delete_stateful_set")
@patch("flytekitplugins.k8sdataservice.connector.K8sManager.delete_service")
def test_gnn_connector_no_configmap(mock_delete_service, mock_delete_stateful_set, mock_check_status, mock_create_data_service):
    connector = ConnectorRegistry.get_connector("dataservicetask")
    task_id, task_metadata, task_config = create_test_setup(original_name="gnn-2345", existing_release_name="gnn-2345")

    int_type = types.LiteralType(types.SimpleType.INTEGER)
    interfaces = interface_models.TypedInterface(
        {
            "a": interface_models.Variable(int_type, "description1"),
            "b": interface_models.Variable(int_type, "description2"),
        },
        {},
    )
    task_inputs = literals.LiteralMap(
        {
            "a": literals.Literal(scalar=literals.Scalar(primitive=literals.Primitive(integer=1))),
            "b": literals.Literal(scalar=literals.Scalar(primitive=literals.Primitive(integer=1))),
        },
    )

    dummy_template = TaskTemplate(
        id=task_id,
        custom=task_config,
        metadata=task_metadata,
        interface=interfaces,
        type="dataservicetask",
    )

    expected_resource_metadata = DataServiceMetadata(
            dataservice_config=DataServiceConfig(
                Name="gnn-2345", Image="image", Command=cmd, Cluster="ei-dev2", ExistingReleaseName="gnn-2345"),
            name="gnn-2345")

    # Test create method, and create_data_service should have not been called
    res_resource_metadata = connector.create(dummy_template, task_inputs)
    assert res_resource_metadata == expected_resource_metadata
    mock_create_data_service.assert_not_called()

    # Test get method
    res = connector.get(res_resource_metadata)
    assert res.phase == TaskExecution.SUCCEEDED
    mock_check_status.assert_called_once_with("gnn-2345")

    # # Test delete methods are not called
    connector.delete(res_resource_metadata)
    mock_delete_stateful_set.assert_called_once_with("gnn-2345")
    mock_delete_service.assert_called_once_with("gnn-2345")


@patch("flytekitplugins.k8sdataservice.connector.K8sManager.create_data_service", return_value="gnn-1234")
@patch("flytekitplugins.k8sdataservice.connector.K8sManager.check_stateful_set_status", return_value="pending")
@patch("flytekitplugins.k8sdataservice.connector.K8sManager.delete_stateful_set")
@patch("flytekitplugins.k8sdataservice.connector.K8sManager.delete_service")
def test_gnn_connector_status_failed(mock_delete_service, mock_delete_stateful_set, mock_check_status, mock_create_data_service):
    connector = ConnectorRegistry.get_connector("dataservicetask")
    task_id, task_metadata, task_config = create_test_setup(original_name="gnn-2345", existing_release_name="gnn-2345")

    int_type = types.LiteralType(types.SimpleType.INTEGER)
    interfaces = interface_models.TypedInterface(
        {
            "a": interface_models.Variable(int_type, "description1"),
            "b": interface_models.Variable(int_type, "description2"),
        },
        {},
    )
    task_inputs = literals.LiteralMap(
        {
            "a": literals.Literal(scalar=literals.Scalar(primitive=literals.Primitive(integer=1))),
            "b": literals.Literal(scalar=literals.Scalar(primitive=literals.Primitive(integer=1))),
        },
    )

    dummy_template = TaskTemplate(
        id=task_id,
        custom=task_config,
        metadata=task_metadata,
        interface=interfaces,
        type="dataservicetask",
    )

    expected_resource_metadata = DataServiceMetadata(
            dataservice_config=DataServiceConfig(
                Name="gnn-2345", Image="image", Command=cmd, Cluster="ei-dev2", ExistingReleaseName="gnn-2345"),
            name="gnn-2345")

    # Test create method, and create_data_service should have not been called
    res_resource_metadata = connector.create(dummy_template, task_inputs)
    assert res_resource_metadata == expected_resource_metadata
    mock_create_data_service.assert_not_called()

    # Test get method
    res = connector.get(res_resource_metadata)
    assert res.phase == TaskExecution.RUNNING
    mock_check_status.assert_called_once_with("gnn-2345")

    mock_check_status.return_value = "failed"
    res.phase == TaskExecution.FAILED

    # # Test delete methods are not called
    connector.delete(res_resource_metadata)
    mock_delete_stateful_set.assert_called_once_with("gnn-2345")
    mock_delete_service.assert_called_once_with("gnn-2345")



from dataclasses import asdict, dataclass
from typing import Any, Dict, List, Optional, Type

from google.protobuf import json_format
from google.protobuf.struct_pb2 import Struct

from flytekit import Resources, kwtypes, logger
from flytekit.configuration import SerializationSettings
from flytekit.core.base_task import PythonTask
from flytekit.core.interface import Interface
from flytekit.extend.backend.base_connector import AsyncConnectorExecutorMixin


@dataclass
class DataServiceConfig(object):
    """DataServiceConfig should be used to configure a DataServiceTask."""

    Name: Optional[str] = None
    Requests: Optional[Resources] = None
    Limits: Optional[Resources] = None
    Port: Optional[int] = None
    Image: Optional[str] = None
    Command: Optional[List[str]] = None
    Replicas: Optional[int] = None
    ExistingReleaseName: Optional[str] = None
    Cluster: Optional[str] = None


class DataServiceTask(AsyncConnectorExecutorMixin, PythonTask[DataServiceConfig]):
    _TASK_TYPE = "dataservicetask"

    def __init__(
        self,
        name: str,
        task_config: Optional[DataServiceConfig],
        inputs: Optional[Dict[str, Type]] = None,
        **kwargs,
    ):
        super().__init__(
            name=name,
            task_config=task_config,
            interface=Interface(inputs=inputs, outputs=kwtypes(name=str)),
            task_type=self._TASK_TYPE,
            **kwargs,
        )

    def get_custom(self, settings: SerializationSettings) -> Dict[str, Any]:
        logger.info("get_custom is invoked")
        config = {}
        limits = None
        requests = None
        if self.task_config is not None:
            limits = asdict(self.task_config.Limits) if self.task_config.Limits is not None else None
            requests = asdict(self.task_config.Requests) if self.task_config.Requests is not None else None
            ge = {
                "Name": self.task_config.Name,
                "Image": self.task_config.Image,
                "Command": self.task_config.Command,
                "Port": self.task_config.Port,
                "Replicas": self.task_config.Replicas,
                "ExistingReleaseName": self.task_config.ExistingReleaseName,
                "Cluster": self.task_config.Cluster,
            }
            if limits is not None:
                ge["Limits"] = limits
            if requests is not None:
                ge["Requests"] = requests
            config = ge
        s = Struct()
        s.update(config)
        return json_format.MessageToDict(s)



from flytekitplugins.k8sdataservice.k8s.kube_config import KubeConfig
from kubernetes import client
from kubernetes.client.rest import ApiException

from flytekit import logger
from flytekit.sensor.base_sensor import BaseSensor

TRAININGJOB_API_GROUP = "kubeflow.org"
VERSION = "v1"


class CleanupSensor(BaseSensor):
    def __init__(self, name: str, namespace: str = "flyte", **kwargs):
        """
        Initialize the CleanupSensor class with relevant configurations for monitoring and managing the k8s data service.
        """
        super().__init__(name=name, task_type="sensor", **kwargs)
        self.k8s_config = KubeConfig()
        try:
            self.k8s_config.load_kube_config()
        except kubernetes.config.ConfigException as e:
            logger.error(f"Failed to load kubernetes config: {e}")
            raise
        self.apps_v1_api = client.AppsV1Api()
        self.core_v1_api = client.CoreV1Api()
        self.custom_api = client.CustomObjectsApi()
        self.namespace = namespace

    async def poke(self, release_name: str, cleanup_data_service: bool, cluster: str) -> bool:
        """poke will delete the graph engine resources based on the user's configuration
        1. This has to be done in the control plane by design. We don't expect any users's running pod to be authn/z to manage resources
        2. This can not be done in the async connector because the delete callback is only invoked on abortion operation or failure phase.
           while this makes sense but what we need is a separate task to delete graph engine without complicating the regular async connector flow.
        3. In the near future, we will add the poking logic on the training job's status. In the initial implementation, we skipped
        it for simplicity. This is also why we use the sensor API to keep forward compatibility
        """
        self.release_name = release_name
        self.cleanup_data_service = cleanup_data_service
        self.cluster = cluster
        return await self._handle_cleanup()

    async def _handle_cleanup(self) -> bool:
        if not self.cleanup_data_service:
            logger.info(
                f"User decides to not to clean up the graph engine: {self.release_name} in cluster {self.cluster}, namespace {self.namespace}"
            )
            logger.info("DataService sensor will stop polling")
            return True
        logger.info(f"The training job is in terminal stage, deleting graph engine {self.release_name}")
        self.delete_data_service()
        return True

    def delete_data_service(self):
        """
        Delete the data service's associated Kubernetes resources (StatefulSet and Service).
        """

        def delete_resource(resource_type: str, delete_fn):
            try:
                delete_fn(name=self.release_name, namespace=self.namespace, body=client.V1DeleteOptions())
                logger.info(f"Deleted {resource_type}: {self.release_name}")
            except ApiException as e:
                logger.error(f"Error deleting {resource_type}: {e}")

        logger.info(f"Sensor got the release name: {self.release_name}")
        delete_resource("Service", self.core_v1_api.delete_namespaced_service)
        delete_resource("StatefulSet", self.apps_v1_api.delete_namespaced_stateful_set)
